/*
 * This program receives some data through a channel of an SFPDP card. 
 * It has two threads: one is to gain the data from the channel and to fill one of ring buffers, 
 * the other is to read the data with the specified size from the ring buffer and to return to user
 * when a synchronized signal (interrupt) is received.
 * If the data buffer is empty, it is waiting for receiving data. 
 * If some new data is received, the above routine is repeated again.
 * At 13 May 2021, it was modifid by Jeongki Kim.
 *
 */
#include <stdio.h>
#include <stdlib.h>		// atoi()
#include <pthread.h>
#include <string.h>		// memcpy()
#include <unistd.h>		// read, write
#include <fcntl.h>		// open, close, O_RDWR, O_NONBLOCK
#include <time.h>
#include <semaphore.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/poll.h>

#include "gefcmn_gpio.h"
#include "sfpdp-api.h"

extern int scGpioPinInterptInit();
// extern int scGpioInit();
extern int scGpioClose();
extern int scGpioSetHigh7();
extern int scGpioSetLow7();
extern GEF_GPIO_HDL gpioHdl;

#define NUM_RXBUF	1024

struct data_buf {
	int	      len;
	unsigned char rdata[RX_FRMSIZE+16];
};

int RXsize=4096;
sem_t rxSem;

struct data_buf RXBuf[NUM_RXBUF];
int hIDX =0;			// header index of RXBuf, which is a location to receive data. 
int rIDX =NUM_RXBUF-1;		// rear index of RXBuf, which is a location to return data to user.



void *gpio_thread(void* arg)		// arg : Intrpt GPIO input pin #: 0 is MCP?
{
	GEF_GPIO_PIN_BITMAP pinBitmap;
	int i;
	int gnum = *(int *)arg;

#ifdef _DEBUG_
	GEF_STATUS gret;
	struct timeval timeout;

	timeout.tv_sec = 1;     	// timeout (seconds)
	timeout.tv_usec = 0;
#endif
	pinBitmap = (0x01 << gnum );

	while(1) {

		for(i=0; i<1; i++) {		// Count # of GPIO interpt.
#ifdef _DEBUG_
			printf("GPIO Pinmap: 0x%08X.\n", pinBitmap );
			gret = gefGpioWaitForInterrupt( gpioHdl, &pinBitmap, &timeout );
			if( gret == GEF_STATUS_SUCCESS )
				printf("GPIO(pin#%d) Interrupted..\n", gnum );
			if( gret == GEF_STATUS_TIMED_OUT )
				printf("GPIO(pin#%d) Timed out..\n", gnum );
#else
			gefGpioWaitForInterrupt( gpioHdl, &pinBitmap, NULL );
#endif
		}
		sem_post( &rxSem );
	}
	return NULL;
}


int increase_RXBuf_index()
{
	rIDX ++;
	if( rIDX == NUM_RXBUF ) rIDX =0;

	while( rIDX == hIDX ) {
		printf("RXBuf buffer is empty (hIDX:%d, rIDX:%d).. Waiting for 1 sec.\n", hIDX, rIDX );
		sleep(1);
	}
	return 0;
}



void *rtndata_thread(void* arg)
{
	int reqsize = *(int *)arg;	// requested size to return to user.
	int rtnsize;			// size of a return block to fill the data.
	int rtnidx =0;			// current data index location of retrun block.
	int didx =0;			// current data index of rIDX-th block among ring buffers.
	unsigned char rtnbuf[RX_BUFSIZE+16];

	rIDX = NUM_RXBUF-1;
	increase_RXBuf_index();

	while(1) {				// to return one block requested by an user.

		sem_wait( &rxSem );		// waiting for GPIO interpt.
		rtnsize = reqsize;
		rtnidx =0;

		while( RXBuf[rIDX].len < rtnsize ) {	// to merge into a return block from rx buffers.

			memcpy( (void*)&rtnbuf[rtnidx], (void*)&(RXBuf[rIDX].rdata[didx]), RXBuf[rIDX].len );
			rtnidx  += RXBuf[rIDX].len;
			rtnsize -= RXBuf[rIDX].len;
#ifdef _DEBUG_
			printf( "Curr rtnidx:%d, rtnsize:%d, didx:%d, RXBUf[%d].len:%d.\n", 
					rtnidx, rtnsize, didx, rIDX, RXBuf[rIDX].len );	
#endif
			increase_RXBuf_index();
			didx =0;
		}

		memcpy( (void*)&rtnbuf[rtnidx], (void*)&(RXBuf[rIDX].rdata[didx]), rtnsize );
		rtnidx  += rtnsize; 
		didx += rtnsize; 
		RXBuf[rIDX].len -= rtnsize;
#ifdef _DEBUG_
		print_buff( rtnbuf, reqsize );	
		printf( "rtnsize:%d, rtnidx:%d, didx: %d RXBUf[%d].len:%d.\n", 
				rtnsize, rtnidx, didx, rIDX, RXBuf[rIDX].len );
#endif
	}
}


/* This receives data from sFPDP channel and fulfill to one of the ring buffer RXBuf. */
void *recv_thread(void* arg)
{
	int ch=0, i;
	int nIndex;
	int retlen; 
	int totrecv=0;
	unsigned int	chkdead, rxchksum;
//	unsigned char	rbuf[RX_BUFSIZE+16];
	struct timeval mytime; 

	ch = *(int *)arg;
	hIDX = 0;

	while(1) {

		retlen = sfpdp_recv_single( ch, (char *)RXBuf[hIDX].rdata, RX_FRMSIZE );
		if( retlen<0 ) {
			printf("\nERR: channel link RX err, retlen: %d\n", retlen );
			return NULL; 
		}

		memcpy( (void*)&chkdead, (void*)&(RXBuf[hIDX].rdata[retlen-sizeof(int)]), sizeof(int) );
		if( chkdead == 0xDEADBEEF )	retlen -= sizeof(int);
		RXBuf[hIDX].len = retlen; 

//		scGpioSetHigh7();
#ifdef _DEBUG_
		printf("\nCH[%d] RXBuf[%d].len: %d Bytes\n", ch, hIDX, retlen );
//		printf("rbuf[0] =0x%02X, rbuf[1] =0x%02X ", rbuf[0], rbuf[1] );
#endif
		// to calcurate check-sum of RX data
		rxchksum = 0;
		for( i=0; i<(retlen-sizeof(int)); i++ ) rxchksum += RXBuf[hIDX].rdata[i];
		
		totrecv += retlen; 
//		scGpioSetLow7();

		hIDX++;
		if( hIDX == NUM_RXBUF ) hIDX =0;
		while( hIDX == rIDX ) {
			printf("RXBuf buffer is full.. Waiting for 1 sec.\n");
			sleep(1);
		}

#ifdef _DEBUG_
		printf("Ch[%d] RX Total size: %d Bytes\n", ch, totrecv );
		printf("Calculated RX chksum: 0x%X.\n", rxchksum ); 
#endif
	}

	return NULL;
}



int init_semas()
{
	if( sem_init( &rxSem, 0, 0) == -1) {
		printf("[%s] Semaphore Init ERROR.\n", __FUNCTION__);
		return -1;
	}
}


int destroy_semas()
{
	sem_destroy( &rxSem );
}



int get_main_arg( int argc, char **argv )
{
	int chno;

	if( argc != 3 ) printf( "Usage: %s [ch#] [RXsize]\n\n", argv[0] );
	if( argc > 3 ) return -1;

	while(1) {
		if( argc == 1 ) {
			printf( "Select the channel# and size to recieve data..\n" );
			printf( "Range of the Channel number is from 0 to %d: ", NUM_OF_DMA );
			scanf( "%d", &chno );
			printf( "Size of RX data (16 ~ %d Byte): ", TX_BUFSIZE );
			scanf( "%d", &RXsize );
		}
		else if( argc == 2 ) {
			chno = atoi( argv[1] );
			printf( "Size of Tx data (16 ~ %d Byte): ", TX_BUFSIZE );
			scanf( "%d", &RXsize );
		}
		else if( argc == 3 ) {
			chno = atoi( argv[1] ); 
			RXsize = atoi( argv[2] ); 
		}

		if( (chno<0) || (chno>NUM_OF_DMA) || (RXsize<16) || (RXsize >TX_BUFSIZE) ) {
			printf( "Range of Input is over: Ch[%d], size:%d\n", chno, RXsize );
			argc = 1;
			continue; 
		}
		break; 
	}
	return chno;
}



int main(int argc, char **argv)
{
	int semIndex = 0;
	int chno, gpnum=0;
	int igpioID, irecvID, irtnID;
	pthread_t p_gpio;
	pthread_t p_recv;
	pthread_t p_rtn;

	chno = get_main_arg( argc, argv );

	printf( "\nExecute: %s ch[%d] %d(RXsize)..\n", argv[0], chno, RXsize );

	/* First, GPIO initialize */
	scGpioPinInterptInit( gpnum );
	init_semas();

	if( sfpdp_ch_open(chno) != 0 )		return -1;
	sfpdp_bdbuf_clear(chno);

	/* 0x01000000 mode means "after the match pattern, data is recieved".
	 * In sfpdp device driver, it was already adapted. */
//	sfpdp_init_control( chno, 0x01000000 );

	printf("SFPDP RX start ..\n");

	igpioID = pthread_create( &p_gpio, NULL, gpio_thread, (void *)&gpnum );
	irtnID  = pthread_create( &p_rtn,  NULL, rtndata_thread, (void *)&RXsize );
	irecvID = pthread_create( &p_recv, NULL, recv_thread, (void *)&chno );

	pthread_join(p_recv, NULL);
	pthread_join(p_rtn,  NULL);
	pthread_join(p_gpio, NULL);

	/* Finalize */
	scGpioClose();
	destroy_semas();
	sfpdp_ch_close(chno);

	return 0;
}

