/*
 * send message through a channel of an SFPDP Card.
 * At 10 May 2021, it was modified by Jeongki Kim.
 */

#include <stdio.h>
#include <stdlib.h>		// atoi()
#include <pthread.h>
#include <string.h>		// memcpy()
#include <unistd.h>		// read, write, usleep()
#include <fcntl.h>		// open, close, O_RDWR, O_NONBLOCK
#include <time.h>
#include <semaphore.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/poll.h>

#include "gefcmn_gpio.h"
#include "sfpdp-api.h"


extern int scGpioPinInterptInit();
// extern int scGpioInit();
extern int scGpioSet();
extern int scGpioClose();
extern int scGpioSetHigh6();
extern int scGpioSetLow6();

extern GEF_GPIO_HDL gpioHdl;

static int TXsize=4096;

sem_t sendSem;


void *gpio_thread( void* arg )		// arg : Intrpt GPIO input pin #: 0 is MCP?
{
	GEF_GPIO_PIN_BITMAP pinBitmap;
	int i;
	int gnum = *(int *)arg;

#ifdef _DEBUG_
	GEF_STATUS gret; 
	struct timeval timeout;

	timeout.tv_sec = 2;	// timeout (seconds)
	timeout.tv_usec = 0;
#endif

	pinBitmap = (0x0001 << gnum );

	while(1) {
		getchar();

		for(i=0; i<1; i++) {
#ifdef _DEBUG_
			printf("GPIO Pinmap: 0x%08X.\n", pinBitmap );
			gret = gefGpioWaitForInterrupt( gpioHdl, &pinBitmap, &timeout );
			if( gret == GEF_STATUS_SUCCESS ) 
				printf("GPIO(pin#%d) Interrupted..\n", gnum );
			if( gret == GEF_STATUS_TIMED_OUT ) 
				printf("GPIO(pin#%d) Timed out..\n", gnum );
#else
			gefGpioWaitForInterrupt( gpioHdl, &pinBitmap, NULL );
#endif
		}

		sem_post( &sendSem );
	}
	return NULL;
}



/* It sends data with TXsize through the SFPDP ch (="arg"). */
void *send_thread(void* arg)
{
	int ch=0;
	int i, rtn=0, cnt = 0, totsend=0;		// total size of send
	unsigned int	match_pattern = 0x5555AAAA;
        unsigned int	rever_pattern = 0xAAAA5555;
	unsigned int	chksum=0;			// for packet data integrity 
	unsigned char	tbuf[TX_BUFSIZE+64];		// 16 KB
	int count = 0;
	struct timeval mytime; 

	ch = *(int *)arg;
	srand( time(NULL) );

	while(1) {
		chksum = 0;
		memcpy( (void*)&tbuf[0], (void*)&match_pattern, sizeof(int) );
		memcpy( (void*)&tbuf[4], (void*)&match_pattern, sizeof(int) );
		memcpy( (void*)&tbuf[8], (void*)&rever_pattern, sizeof(int) );
		memcpy( (void*)&tbuf[12], (void*)&rever_pattern, sizeof(int) );
		tbuf[16] =cnt;

		for(i=17; i < (TXsize-sizeof(int)); i++) {
//			tbuf[i] = rand() % 10;
			tbuf[i] = i % 256;
		}

		for(i=0; i < (TXsize-sizeof(int)); i++)  chksum += tbuf[i];
		memcpy( (void*)&tbuf[i], (void*)&chksum, sizeof(int) );

		sem_wait( &sendSem );
		scGpioSetHigh6();

		rtn = sfpdp_send_with_swdv( ch, (char *)tbuf, TXsize );
		gettimeofday( &mytime, NULL );
		printf( "\n Curr time (usec) TX_user: %ld \n", mytime.tv_usec );

#ifdef _DEBUG_
		if( !(rtn>0) ) {
			printf( "%d-th block, there is no data of sfpdp_send.\n", cnt );
		}
		print_buff( tbuf, TXsize );
		printf( "CH[%d] real TX size:%d Bytes (Req Tx_size:%d)\n", ch, rtn, TXsize );
#endif
		totsend += rtn;
		cnt++;

		scGpioSetLow6();
#ifdef _DEBUG_
		printf("Ch[%d] Total TX Size: %d Bytes\n", ch, totsend );
		printf("Ch[%d] TX Check-sum value: 0x%X\n", ch, chksum );
#endif
	} // while

	return NULL;
}



int get_main_args( int argc, char **argv )
{
	int chno;

	if( argc != 3 ) printf( "Usage: %s [ch#] [TXsize]\n\n", argv[0] );
	if( argc > 3 ) return -1;

	while(1) {
		if( argc == 1 ) {
			printf( "Select the channel and size to transfer data..\n" );
			printf( "Range of the Channel number is from 0 to %d: ", NUM_OF_DMA );
			scanf( "%d", &chno );

			printf( "Size of Tx data (16 ~ %d Byte): ", TX_BUFSIZE );
			scanf( "%d", &TXsize );
		}
		else if( argc == 2 ) {
			chno = atoi( argv[1] );
			printf( "Size of Tx data (16 ~ %d Byte): ", TX_BUFSIZE );
			scanf( "%d", &TXsize );
		}
		else if( argc == 3 ) {
			chno = atoi( argv[1] ); 
			TXsize = atoi( argv[2] ); 
		}

		if( (chno<0) || (chno>NUM_OF_DMA) || (TXsize <16) || (TXsize >TX_BUFSIZE) ) {
			printf( "Range of the data is over: %d\n", TXsize );
			argc =1;
			continue; 
		}
		else break;
	}
	return chno;
}



int init_sema()
{
	if( sem_init( &sendSem, 0, 0) == -1) {
		printf("[%s] Semaphore Init ERROR.\n", __FUNCTION__);
		return -1;
	}
}


int destroy_sema()
{
	sem_destroy( &sendSem );
}



int main(int argc, char **argv)
{
	int gpin =1;		// GPIO interpt pin.
	int chno;

	pthread_t p_gpio;
	pthread_t p_send;

	int igpioID;
	int isendID;

	chno = get_main_args( argc, argv );

	printf( "Execute: %s Ch[%d] %d(TXsize) through SFPDP card..\n", argv[0], chno, TXsize );

	/* First, GPIO initialize */
	scGpioPinInterptInit( gpin );	// using gpio #gnum as interrupt.
	init_sema();

	if( sfpdp_ch_open(chno) != 0 )			return -1;
	sfpdp_bdbuf_clear(chno);

	igpioID = pthread_create( &p_gpio, NULL, gpio_thread, (void *)&gpin );
	isendID = pthread_create( &p_send, NULL, send_thread, (void *)&chno );

//	printf("SFPDP send start ..\n");

	pthread_join(p_send, NULL);
	pthread_join(p_gpio, NULL);

	/* Finalize */
	scGpioClose();
	destroy_sema();

	sfpdp_ch_close(chno);

	return 0;
}


